/* automatically generated by rust-bindgen 0.72.1 */

pub const JS_PROP_CONFIGURABLE: u32 = 1;
pub const JS_PROP_WRITABLE: u32 = 2;
pub const JS_PROP_ENUMERABLE: u32 = 4;
pub const JS_PROP_C_W_E: u32 = 7;
pub const JS_PROP_LENGTH: u32 = 8;
pub const JS_PROP_TMASK: u32 = 48;
pub const JS_PROP_NORMAL: u32 = 0;
pub const JS_PROP_GETSET: u32 = 16;
pub const JS_PROP_VARREF: u32 = 32;
pub const JS_PROP_AUTOINIT: u32 = 48;
pub const JS_PROP_HAS_SHIFT: u32 = 8;
pub const JS_PROP_HAS_CONFIGURABLE: u32 = 256;
pub const JS_PROP_HAS_WRITABLE: u32 = 512;
pub const JS_PROP_HAS_ENUMERABLE: u32 = 1024;
pub const JS_PROP_HAS_GET: u32 = 2048;
pub const JS_PROP_HAS_SET: u32 = 4096;
pub const JS_PROP_HAS_VALUE: u32 = 8192;
pub const JS_PROP_THROW: u32 = 16384;
pub const JS_PROP_THROW_STRICT: u32 = 32768;
pub const JS_PROP_NO_ADD: u32 = 65536;
pub const JS_PROP_NO_EXOTIC: u32 = 131072;
pub const JS_PROP_DEFINE_PROPERTY: u32 = 262144;
pub const JS_PROP_REFLECT_DEFINE_PROPERTY: u32 = 524288;
pub const JS_DEFAULT_STACK_SIZE: u32 = 1048576;
pub const JS_EVAL_TYPE_GLOBAL: u32 = 0;
pub const JS_EVAL_TYPE_MODULE: u32 = 1;
pub const JS_EVAL_TYPE_DIRECT: u32 = 2;
pub const JS_EVAL_TYPE_INDIRECT: u32 = 3;
pub const JS_EVAL_TYPE_MASK: u32 = 3;
pub const JS_EVAL_FLAG_STRICT: u32 = 8;
pub const JS_EVAL_FLAG_UNUSED: u32 = 16;
pub const JS_EVAL_FLAG_COMPILE_ONLY: u32 = 32;
pub const JS_EVAL_FLAG_BACKTRACE_BARRIER: u32 = 64;
pub const JS_EVAL_FLAG_ASYNC: u32 = 128;
pub const JS_DUMP_BYTECODE_FINAL: u32 = 1;
pub const JS_DUMP_BYTECODE_PASS2: u32 = 2;
pub const JS_DUMP_BYTECODE_PASS1: u32 = 4;
pub const JS_DUMP_BYTECODE_HEX: u32 = 16;
pub const JS_DUMP_BYTECODE_PC2LINE: u32 = 32;
pub const JS_DUMP_BYTECODE_STACK: u32 = 64;
pub const JS_DUMP_BYTECODE_STEP: u32 = 128;
pub const JS_DUMP_READ_OBJECT: u32 = 256;
pub const JS_DUMP_FREE: u32 = 512;
pub const JS_DUMP_GC: u32 = 1024;
pub const JS_DUMP_GC_FREE: u32 = 2048;
pub const JS_DUMP_MODULE_RESOLVE: u32 = 4096;
pub const JS_DUMP_PROMISE: u32 = 8192;
pub const JS_DUMP_LEAKS: u32 = 16384;
pub const JS_DUMP_ATOM_LEAKS: u32 = 32768;
pub const JS_DUMP_MEM: u32 = 65536;
pub const JS_DUMP_OBJECTS: u32 = 131072;
pub const JS_DUMP_ATOMS: u32 = 262144;
pub const JS_DUMP_SHAPES: u32 = 524288;
pub const JS_ATOM_NULL: u32 = 0;
pub const JS_CALL_FLAG_CONSTRUCTOR: u32 = 1;
pub const JS_EVAL_OPTIONS_VERSION: u32 = 1;
pub const JS_INVALID_CLASS_ID: u32 = 0;
pub const JS_GPN_STRING_MASK: u32 = 1;
pub const JS_GPN_SYMBOL_MASK: u32 = 2;
pub const JS_GPN_PRIVATE_MASK: u32 = 4;
pub const JS_GPN_ENUM_ONLY: u32 = 16;
pub const JS_GPN_SET_ENUM: u32 = 32;
pub const JS_WRITE_OBJ_BYTECODE: u32 = 1;
pub const JS_WRITE_OBJ_BSWAP: u32 = 0;
pub const JS_WRITE_OBJ_SAB: u32 = 4;
pub const JS_WRITE_OBJ_REFERENCE: u32 = 8;
pub const JS_WRITE_OBJ_STRIP_SOURCE: u32 = 16;
pub const JS_WRITE_OBJ_STRIP_DEBUG: u32 = 32;
pub const JS_READ_OBJ_BYTECODE: u32 = 1;
pub const JS_READ_OBJ_ROM_DATA: u32 = 0;
pub const JS_READ_OBJ_SAB: u32 = 4;
pub const JS_READ_OBJ_REFERENCE: u32 = 8;
pub const JS_DEF_CFUNC: u32 = 0;
pub const JS_DEF_CGETSET: u32 = 1;
pub const JS_DEF_CGETSET_MAGIC: u32 = 2;
pub const JS_DEF_PROP_STRING: u32 = 3;
pub const JS_DEF_PROP_INT32: u32 = 4;
pub const JS_DEF_PROP_INT64: u32 = 5;
pub const JS_DEF_PROP_DOUBLE: u32 = 6;
pub const JS_DEF_PROP_UNDEFINED: u32 = 7;
pub const JS_DEF_OBJECT: u32 = 8;
pub const JS_DEF_ALIAS: u32 = 9;
pub type __darwin_size_t = ::core::ffi::c_ulong;
pub type size_t = __darwin_size_t;
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSRuntime {
    _unused: [u8; 0],
}
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSContext {
    _unused: [u8; 0],
}
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSObject {
    _unused: [u8; 0],
}
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSClass {
    _unused: [u8; 0],
}
pub type JSClassID = u32;
pub type JSAtom = u32;
pub const JS_TAG_FIRST: _bindgen_ty_1 = -9;
pub const JS_TAG_BIG_INT: _bindgen_ty_1 = -9;
pub const JS_TAG_SYMBOL: _bindgen_ty_1 = -8;
pub const JS_TAG_STRING: _bindgen_ty_1 = -7;
pub const JS_TAG_MODULE: _bindgen_ty_1 = -3;
pub const JS_TAG_FUNCTION_BYTECODE: _bindgen_ty_1 = -2;
pub const JS_TAG_OBJECT: _bindgen_ty_1 = -1;
pub const JS_TAG_INT: _bindgen_ty_1 = 0;
pub const JS_TAG_BOOL: _bindgen_ty_1 = 1;
pub const JS_TAG_NULL: _bindgen_ty_1 = 2;
pub const JS_TAG_UNDEFINED: _bindgen_ty_1 = 3;
pub const JS_TAG_UNINITIALIZED: _bindgen_ty_1 = 4;
pub const JS_TAG_CATCH_OFFSET: _bindgen_ty_1 = 5;
pub const JS_TAG_EXCEPTION: _bindgen_ty_1 = 6;
pub const JS_TAG_SHORT_BIG_INT: _bindgen_ty_1 = 7;
pub const JS_TAG_FLOAT64: _bindgen_ty_1 = 8;
pub type _bindgen_ty_1 = ::core::ffi::c_int;
#[repr(C)]
#[derive(Copy, Clone)]
pub union JSValueUnion {
    pub int32: i32,
    pub float64: f64,
    pub ptr: *mut ::core::ffi::c_void,
    pub short_big_int: i32,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSValueUnion"][::core::mem::size_of::<JSValueUnion>() - 8usize];
    ["Alignment of JSValueUnion"][::core::mem::align_of::<JSValueUnion>() - 8usize];
    ["Offset of field: JSValueUnion::int32"][::core::mem::offset_of!(JSValueUnion, int32) - 0usize];
    ["Offset of field: JSValueUnion::float64"]
        [::core::mem::offset_of!(JSValueUnion, float64) - 0usize];
    ["Offset of field: JSValueUnion::ptr"][::core::mem::offset_of!(JSValueUnion, ptr) - 0usize];
    ["Offset of field: JSValueUnion::short_big_int"]
        [::core::mem::offset_of!(JSValueUnion, short_big_int) - 0usize];
};
#[repr(C)]
#[derive(Copy, Clone)]
pub struct JSValue {
    pub u: JSValueUnion,
    pub tag: i64,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSValue"][::core::mem::size_of::<JSValue>() - 16usize];
    ["Alignment of JSValue"][::core::mem::align_of::<JSValue>() - 8usize];
    ["Offset of field: JSValue::u"][::core::mem::offset_of!(JSValue, u) - 0usize];
    ["Offset of field: JSValue::tag"][::core::mem::offset_of!(JSValue, tag) - 8usize];
};
pub type JSCFunction = ::core::option::Option<
    unsafe extern "C" fn(
        ctx: *mut JSContext,
        this_val: JSValue,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
    ) -> JSValue,
>;
pub type JSCFunctionMagic = ::core::option::Option<
    unsafe extern "C" fn(
        ctx: *mut JSContext,
        this_val: JSValue,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
        magic: ::core::ffi::c_int,
    ) -> JSValue,
>;
pub type JSCFunctionData = ::core::option::Option<
    unsafe extern "C" fn(
        ctx: *mut JSContext,
        this_val: JSValue,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
        magic: ::core::ffi::c_int,
        func_data: *mut JSValue,
    ) -> JSValue,
>;
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSMallocFunctions {
    pub js_calloc: ::core::option::Option<
        unsafe extern "C" fn(
            opaque: *mut ::core::ffi::c_void,
            count: size_t,
            size: size_t,
        ) -> *mut ::core::ffi::c_void,
    >,
    pub js_malloc: ::core::option::Option<
        unsafe extern "C" fn(
            opaque: *mut ::core::ffi::c_void,
            size: size_t,
        ) -> *mut ::core::ffi::c_void,
    >,
    pub js_free: ::core::option::Option<
        unsafe extern "C" fn(opaque: *mut ::core::ffi::c_void, ptr: *mut ::core::ffi::c_void),
    >,
    pub js_realloc: ::core::option::Option<
        unsafe extern "C" fn(
            opaque: *mut ::core::ffi::c_void,
            ptr: *mut ::core::ffi::c_void,
            size: size_t,
        ) -> *mut ::core::ffi::c_void,
    >,
    pub js_malloc_usable_size:
        ::core::option::Option<unsafe extern "C" fn(ptr: *const ::core::ffi::c_void) -> size_t>,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSMallocFunctions"][::core::mem::size_of::<JSMallocFunctions>() - 40usize];
    ["Alignment of JSMallocFunctions"][::core::mem::align_of::<JSMallocFunctions>() - 8usize];
    ["Offset of field: JSMallocFunctions::js_calloc"]
        [::core::mem::offset_of!(JSMallocFunctions, js_calloc) - 0usize];
    ["Offset of field: JSMallocFunctions::js_malloc"]
        [::core::mem::offset_of!(JSMallocFunctions, js_malloc) - 8usize];
    ["Offset of field: JSMallocFunctions::js_free"]
        [::core::mem::offset_of!(JSMallocFunctions, js_free) - 16usize];
    ["Offset of field: JSMallocFunctions::js_realloc"]
        [::core::mem::offset_of!(JSMallocFunctions, js_realloc) - 24usize];
    ["Offset of field: JSMallocFunctions::js_malloc_usable_size"]
        [::core::mem::offset_of!(JSMallocFunctions, js_malloc_usable_size) - 32usize];
};
pub type JSRuntimeFinalizer =
    ::core::option::Option<unsafe extern "C" fn(rt: *mut JSRuntime, arg: *mut ::core::ffi::c_void)>;
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSGCObjectHeader {
    _unused: [u8; 0],
}
unsafe extern "C" {
    pub fn JS_NewRuntime() -> *mut JSRuntime;
}
unsafe extern "C" {
    pub fn JS_SetRuntimeInfo(rt: *mut JSRuntime, info: *const ::core::ffi::c_char);
}
unsafe extern "C" {
    pub fn JS_SetMemoryLimit(rt: *mut JSRuntime, limit: size_t);
}
unsafe extern "C" {
    pub fn JS_SetDumpFlags(rt: *mut JSRuntime, flags: u64);
}
unsafe extern "C" {
    pub fn JS_GetDumpFlags(rt: *mut JSRuntime) -> u64;
}
unsafe extern "C" {
    pub fn JS_GetGCThreshold(rt: *mut JSRuntime) -> size_t;
}
unsafe extern "C" {
    pub fn JS_SetGCThreshold(rt: *mut JSRuntime, gc_threshold: size_t);
}
unsafe extern "C" {
    pub fn JS_SetMaxStackSize(rt: *mut JSRuntime, stack_size: size_t);
}
unsafe extern "C" {
    pub fn JS_UpdateStackTop(rt: *mut JSRuntime);
}
unsafe extern "C" {
    pub fn JS_NewRuntime2(
        mf: *const JSMallocFunctions,
        opaque: *mut ::core::ffi::c_void,
    ) -> *mut JSRuntime;
}
unsafe extern "C" {
    pub fn JS_FreeRuntime(rt: *mut JSRuntime);
}
unsafe extern "C" {
    pub fn JS_GetRuntimeOpaque(rt: *mut JSRuntime) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn JS_SetRuntimeOpaque(rt: *mut JSRuntime, opaque: *mut ::core::ffi::c_void);
}
unsafe extern "C" {
    pub fn JS_AddRuntimeFinalizer(
        rt: *mut JSRuntime,
        finalizer: JSRuntimeFinalizer,
        arg: *mut ::core::ffi::c_void,
    ) -> ::core::ffi::c_int;
}
pub type JS_MarkFunc =
    ::core::option::Option<unsafe extern "C" fn(rt: *mut JSRuntime, gp: *mut JSGCObjectHeader)>;
unsafe extern "C" {
    pub fn JS_MarkValue(rt: *mut JSRuntime, val: JSValue, mark_func: JS_MarkFunc);
}
unsafe extern "C" {
    pub fn JS_RunGC(rt: *mut JSRuntime);
}
unsafe extern "C" {
    pub fn JS_IsLiveObject(rt: *mut JSRuntime, obj: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_NewContext(rt: *mut JSRuntime) -> *mut JSContext;
}
unsafe extern "C" {
    pub fn JS_FreeContext(s: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_DupContext(ctx: *mut JSContext) -> *mut JSContext;
}
unsafe extern "C" {
    pub fn JS_GetContextOpaque(ctx: *mut JSContext) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn JS_SetContextOpaque(ctx: *mut JSContext, opaque: *mut ::core::ffi::c_void);
}
unsafe extern "C" {
    pub fn JS_GetRuntime(ctx: *mut JSContext) -> *mut JSRuntime;
}
unsafe extern "C" {
    pub fn JS_SetClassProto(ctx: *mut JSContext, class_id: JSClassID, obj: JSValue);
}
unsafe extern "C" {
    pub fn JS_GetClassProto(ctx: *mut JSContext, class_id: JSClassID) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetFunctionProto(ctx: *mut JSContext) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewContextRaw(rt: *mut JSRuntime) -> *mut JSContext;
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicBaseObjects(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicDate(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicEval(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicRegExpCompiler(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicRegExp(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicJSON(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicProxy(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicMapSet(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicTypedArrays(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicPromise(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicBigInt(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicWeakRef(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddPerformance(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_AddIntrinsicDOMException(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_IsEqual(ctx: *mut JSContext, op1: JSValue, op2: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_IsStrictEqual(ctx: *mut JSContext, op1: JSValue, op2: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsSameValue(ctx: *mut JSContext, op1: JSValue, op2: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsSameValueZero(ctx: *mut JSContext, op1: JSValue, op2: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn js_string_codePointRange(
        ctx: *mut JSContext,
        this_val: JSValue,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn js_calloc_rt(
        rt: *mut JSRuntime,
        count: size_t,
        size: size_t,
    ) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn js_malloc_rt(rt: *mut JSRuntime, size: size_t) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn js_free_rt(rt: *mut JSRuntime, ptr: *mut ::core::ffi::c_void);
}
unsafe extern "C" {
    pub fn js_realloc_rt(
        rt: *mut JSRuntime,
        ptr: *mut ::core::ffi::c_void,
        size: size_t,
    ) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn js_malloc_usable_size_rt(rt: *mut JSRuntime, ptr: *const ::core::ffi::c_void) -> size_t;
}
unsafe extern "C" {
    pub fn js_mallocz_rt(rt: *mut JSRuntime, size: size_t) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn js_calloc(ctx: *mut JSContext, count: size_t, size: size_t) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn js_malloc(ctx: *mut JSContext, size: size_t) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn js_free(ctx: *mut JSContext, ptr: *mut ::core::ffi::c_void);
}
unsafe extern "C" {
    pub fn js_realloc(
        ctx: *mut JSContext,
        ptr: *mut ::core::ffi::c_void,
        size: size_t,
    ) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn js_malloc_usable_size(ctx: *mut JSContext, ptr: *const ::core::ffi::c_void) -> size_t;
}
unsafe extern "C" {
    pub fn js_realloc2(
        ctx: *mut JSContext,
        ptr: *mut ::core::ffi::c_void,
        size: size_t,
        pslack: *mut size_t,
    ) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn js_mallocz(ctx: *mut JSContext, size: size_t) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn js_strdup(
        ctx: *mut JSContext,
        str_: *const ::core::ffi::c_char,
    ) -> *mut ::core::ffi::c_char;
}
unsafe extern "C" {
    pub fn js_strndup(
        ctx: *mut JSContext,
        s: *const ::core::ffi::c_char,
        n: size_t,
    ) -> *mut ::core::ffi::c_char;
}
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSMemoryUsage {
    pub malloc_size: i64,
    pub malloc_limit: i64,
    pub memory_used_size: i64,
    pub malloc_count: i64,
    pub memory_used_count: i64,
    pub atom_count: i64,
    pub atom_size: i64,
    pub str_count: i64,
    pub str_size: i64,
    pub obj_count: i64,
    pub obj_size: i64,
    pub prop_count: i64,
    pub prop_size: i64,
    pub shape_count: i64,
    pub shape_size: i64,
    pub js_func_count: i64,
    pub js_func_size: i64,
    pub js_func_code_size: i64,
    pub js_func_pc2line_count: i64,
    pub js_func_pc2line_size: i64,
    pub c_func_count: i64,
    pub array_count: i64,
    pub fast_array_count: i64,
    pub fast_array_elements: i64,
    pub binary_object_count: i64,
    pub binary_object_size: i64,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSMemoryUsage"][::core::mem::size_of::<JSMemoryUsage>() - 208usize];
    ["Alignment of JSMemoryUsage"][::core::mem::align_of::<JSMemoryUsage>() - 8usize];
    ["Offset of field: JSMemoryUsage::malloc_size"]
        [::core::mem::offset_of!(JSMemoryUsage, malloc_size) - 0usize];
    ["Offset of field: JSMemoryUsage::malloc_limit"]
        [::core::mem::offset_of!(JSMemoryUsage, malloc_limit) - 8usize];
    ["Offset of field: JSMemoryUsage::memory_used_size"]
        [::core::mem::offset_of!(JSMemoryUsage, memory_used_size) - 16usize];
    ["Offset of field: JSMemoryUsage::malloc_count"]
        [::core::mem::offset_of!(JSMemoryUsage, malloc_count) - 24usize];
    ["Offset of field: JSMemoryUsage::memory_used_count"]
        [::core::mem::offset_of!(JSMemoryUsage, memory_used_count) - 32usize];
    ["Offset of field: JSMemoryUsage::atom_count"]
        [::core::mem::offset_of!(JSMemoryUsage, atom_count) - 40usize];
    ["Offset of field: JSMemoryUsage::atom_size"]
        [::core::mem::offset_of!(JSMemoryUsage, atom_size) - 48usize];
    ["Offset of field: JSMemoryUsage::str_count"]
        [::core::mem::offset_of!(JSMemoryUsage, str_count) - 56usize];
    ["Offset of field: JSMemoryUsage::str_size"]
        [::core::mem::offset_of!(JSMemoryUsage, str_size) - 64usize];
    ["Offset of field: JSMemoryUsage::obj_count"]
        [::core::mem::offset_of!(JSMemoryUsage, obj_count) - 72usize];
    ["Offset of field: JSMemoryUsage::obj_size"]
        [::core::mem::offset_of!(JSMemoryUsage, obj_size) - 80usize];
    ["Offset of field: JSMemoryUsage::prop_count"]
        [::core::mem::offset_of!(JSMemoryUsage, prop_count) - 88usize];
    ["Offset of field: JSMemoryUsage::prop_size"]
        [::core::mem::offset_of!(JSMemoryUsage, prop_size) - 96usize];
    ["Offset of field: JSMemoryUsage::shape_count"]
        [::core::mem::offset_of!(JSMemoryUsage, shape_count) - 104usize];
    ["Offset of field: JSMemoryUsage::shape_size"]
        [::core::mem::offset_of!(JSMemoryUsage, shape_size) - 112usize];
    ["Offset of field: JSMemoryUsage::js_func_count"]
        [::core::mem::offset_of!(JSMemoryUsage, js_func_count) - 120usize];
    ["Offset of field: JSMemoryUsage::js_func_size"]
        [::core::mem::offset_of!(JSMemoryUsage, js_func_size) - 128usize];
    ["Offset of field: JSMemoryUsage::js_func_code_size"]
        [::core::mem::offset_of!(JSMemoryUsage, js_func_code_size) - 136usize];
    ["Offset of field: JSMemoryUsage::js_func_pc2line_count"]
        [::core::mem::offset_of!(JSMemoryUsage, js_func_pc2line_count) - 144usize];
    ["Offset of field: JSMemoryUsage::js_func_pc2line_size"]
        [::core::mem::offset_of!(JSMemoryUsage, js_func_pc2line_size) - 152usize];
    ["Offset of field: JSMemoryUsage::c_func_count"]
        [::core::mem::offset_of!(JSMemoryUsage, c_func_count) - 160usize];
    ["Offset of field: JSMemoryUsage::array_count"]
        [::core::mem::offset_of!(JSMemoryUsage, array_count) - 168usize];
    ["Offset of field: JSMemoryUsage::fast_array_count"]
        [::core::mem::offset_of!(JSMemoryUsage, fast_array_count) - 176usize];
    ["Offset of field: JSMemoryUsage::fast_array_elements"]
        [::core::mem::offset_of!(JSMemoryUsage, fast_array_elements) - 184usize];
    ["Offset of field: JSMemoryUsage::binary_object_count"]
        [::core::mem::offset_of!(JSMemoryUsage, binary_object_count) - 192usize];
    ["Offset of field: JSMemoryUsage::binary_object_size"]
        [::core::mem::offset_of!(JSMemoryUsage, binary_object_size) - 200usize];
};
unsafe extern "C" {
    pub fn JS_ComputeMemoryUsage(rt: *mut JSRuntime, s: *mut JSMemoryUsage);
}
unsafe extern "C" {
    pub fn JS_NewAtomLen(
        ctx: *mut JSContext,
        str_: *const ::core::ffi::c_char,
        len: size_t,
    ) -> JSAtom;
}
unsafe extern "C" {
    pub fn JS_NewAtom(ctx: *mut JSContext, str_: *const ::core::ffi::c_char) -> JSAtom;
}
unsafe extern "C" {
    pub fn JS_NewAtomUInt32(ctx: *mut JSContext, n: u32) -> JSAtom;
}
unsafe extern "C" {
    pub fn JS_DupAtom(ctx: *mut JSContext, v: JSAtom) -> JSAtom;
}
unsafe extern "C" {
    pub fn JS_FreeAtom(ctx: *mut JSContext, v: JSAtom);
}
unsafe extern "C" {
    pub fn JS_FreeAtomRT(rt: *mut JSRuntime, v: JSAtom);
}
unsafe extern "C" {
    pub fn JS_AtomToValue(ctx: *mut JSContext, atom: JSAtom) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_AtomToString(ctx: *mut JSContext, atom: JSAtom) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_AtomToCStringLen(
        ctx: *mut JSContext,
        plen: *mut size_t,
        atom: JSAtom,
    ) -> *const ::core::ffi::c_char;
}
unsafe extern "C" {
    pub fn JS_ValueToAtom(ctx: *mut JSContext, val: JSValue) -> JSAtom;
}
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSPropertyEnum {
    pub is_enumerable: bool,
    pub atom: JSAtom,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSPropertyEnum"][::core::mem::size_of::<JSPropertyEnum>() - 8usize];
    ["Alignment of JSPropertyEnum"][::core::mem::align_of::<JSPropertyEnum>() - 4usize];
    ["Offset of field: JSPropertyEnum::is_enumerable"]
        [::core::mem::offset_of!(JSPropertyEnum, is_enumerable) - 0usize];
    ["Offset of field: JSPropertyEnum::atom"]
        [::core::mem::offset_of!(JSPropertyEnum, atom) - 4usize];
};
#[repr(C)]
#[derive(Copy, Clone)]
pub struct JSPropertyDescriptor {
    pub flags: ::core::ffi::c_int,
    pub value: JSValue,
    pub getter: JSValue,
    pub setter: JSValue,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSPropertyDescriptor"][::core::mem::size_of::<JSPropertyDescriptor>() - 56usize];
    ["Alignment of JSPropertyDescriptor"][::core::mem::align_of::<JSPropertyDescriptor>() - 8usize];
    ["Offset of field: JSPropertyDescriptor::flags"]
        [::core::mem::offset_of!(JSPropertyDescriptor, flags) - 0usize];
    ["Offset of field: JSPropertyDescriptor::value"]
        [::core::mem::offset_of!(JSPropertyDescriptor, value) - 8usize];
    ["Offset of field: JSPropertyDescriptor::getter"]
        [::core::mem::offset_of!(JSPropertyDescriptor, getter) - 24usize];
    ["Offset of field: JSPropertyDescriptor::setter"]
        [::core::mem::offset_of!(JSPropertyDescriptor, setter) - 40usize];
};
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSClassExoticMethods {
    pub get_own_property: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            desc: *mut JSPropertyDescriptor,
            obj: JSValue,
            prop: JSAtom,
        ) -> ::core::ffi::c_int,
    >,
    pub get_own_property_names: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            ptab: *mut *mut JSPropertyEnum,
            plen: *mut u32,
            obj: JSValue,
        ) -> ::core::ffi::c_int,
    >,
    pub delete_property: ::core::option::Option<
        unsafe extern "C" fn(ctx: *mut JSContext, obj: JSValue, prop: JSAtom) -> ::core::ffi::c_int,
    >,
    pub define_own_property: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            this_obj: JSValue,
            prop: JSAtom,
            val: JSValue,
            getter: JSValue,
            setter: JSValue,
            flags: ::core::ffi::c_int,
        ) -> ::core::ffi::c_int,
    >,
    pub has_property: ::core::option::Option<
        unsafe extern "C" fn(ctx: *mut JSContext, obj: JSValue, atom: JSAtom) -> ::core::ffi::c_int,
    >,
    pub get_property: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            obj: JSValue,
            atom: JSAtom,
            receiver: JSValue,
        ) -> JSValue,
    >,
    pub set_property: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            obj: JSValue,
            atom: JSAtom,
            value: JSValue,
            receiver: JSValue,
            flags: ::core::ffi::c_int,
        ) -> ::core::ffi::c_int,
    >,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSClassExoticMethods"][::core::mem::size_of::<JSClassExoticMethods>() - 56usize];
    ["Alignment of JSClassExoticMethods"][::core::mem::align_of::<JSClassExoticMethods>() - 8usize];
    ["Offset of field: JSClassExoticMethods::get_own_property"]
        [::core::mem::offset_of!(JSClassExoticMethods, get_own_property) - 0usize];
    ["Offset of field: JSClassExoticMethods::get_own_property_names"]
        [::core::mem::offset_of!(JSClassExoticMethods, get_own_property_names) - 8usize];
    ["Offset of field: JSClassExoticMethods::delete_property"]
        [::core::mem::offset_of!(JSClassExoticMethods, delete_property) - 16usize];
    ["Offset of field: JSClassExoticMethods::define_own_property"]
        [::core::mem::offset_of!(JSClassExoticMethods, define_own_property) - 24usize];
    ["Offset of field: JSClassExoticMethods::has_property"]
        [::core::mem::offset_of!(JSClassExoticMethods, has_property) - 32usize];
    ["Offset of field: JSClassExoticMethods::get_property"]
        [::core::mem::offset_of!(JSClassExoticMethods, get_property) - 40usize];
    ["Offset of field: JSClassExoticMethods::set_property"]
        [::core::mem::offset_of!(JSClassExoticMethods, set_property) - 48usize];
};
pub type JSClassFinalizer =
    ::core::option::Option<unsafe extern "C" fn(rt: *mut JSRuntime, val: JSValue)>;
pub type JSClassGCMark = ::core::option::Option<
    unsafe extern "C" fn(rt: *mut JSRuntime, val: JSValue, mark_func: JS_MarkFunc),
>;
pub type JSClassCall = ::core::option::Option<
    unsafe extern "C" fn(
        ctx: *mut JSContext,
        func_obj: JSValue,
        this_val: JSValue,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
        flags: ::core::ffi::c_int,
    ) -> JSValue,
>;
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSClassDef {
    pub class_name: *const ::core::ffi::c_char,
    pub finalizer: JSClassFinalizer,
    pub gc_mark: JSClassGCMark,
    pub call: JSClassCall,
    pub exotic: *mut JSClassExoticMethods,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSClassDef"][::core::mem::size_of::<JSClassDef>() - 40usize];
    ["Alignment of JSClassDef"][::core::mem::align_of::<JSClassDef>() - 8usize];
    ["Offset of field: JSClassDef::class_name"]
        [::core::mem::offset_of!(JSClassDef, class_name) - 0usize];
    ["Offset of field: JSClassDef::finalizer"]
        [::core::mem::offset_of!(JSClassDef, finalizer) - 8usize];
    ["Offset of field: JSClassDef::gc_mark"]
        [::core::mem::offset_of!(JSClassDef, gc_mark) - 16usize];
    ["Offset of field: JSClassDef::call"][::core::mem::offset_of!(JSClassDef, call) - 24usize];
    ["Offset of field: JSClassDef::exotic"][::core::mem::offset_of!(JSClassDef, exotic) - 32usize];
};
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSEvalOptions {
    pub version: ::core::ffi::c_int,
    pub eval_flags: ::core::ffi::c_int,
    pub filename: *const ::core::ffi::c_char,
    pub line_num: ::core::ffi::c_int,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSEvalOptions"][::core::mem::size_of::<JSEvalOptions>() - 24usize];
    ["Alignment of JSEvalOptions"][::core::mem::align_of::<JSEvalOptions>() - 8usize];
    ["Offset of field: JSEvalOptions::version"]
        [::core::mem::offset_of!(JSEvalOptions, version) - 0usize];
    ["Offset of field: JSEvalOptions::eval_flags"]
        [::core::mem::offset_of!(JSEvalOptions, eval_flags) - 4usize];
    ["Offset of field: JSEvalOptions::filename"]
        [::core::mem::offset_of!(JSEvalOptions, filename) - 8usize];
    ["Offset of field: JSEvalOptions::line_num"]
        [::core::mem::offset_of!(JSEvalOptions, line_num) - 16usize];
};
unsafe extern "C" {
    pub fn JS_NewClassID(rt: *mut JSRuntime, pclass_id: *mut JSClassID) -> JSClassID;
}
unsafe extern "C" {
    pub fn JS_GetClassID(v: JSValue) -> JSClassID;
}
unsafe extern "C" {
    pub fn JS_NewClass(
        rt: *mut JSRuntime,
        class_id: JSClassID,
        class_def: *const JSClassDef,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_IsRegisteredClass(rt: *mut JSRuntime, class_id: JSClassID) -> bool;
}
unsafe extern "C" {
    pub fn JS_NewNumber(ctx: *mut JSContext, d: f64) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewBigInt64(ctx: *mut JSContext, v: i64) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewBigUint64(ctx: *mut JSContext, v: u64) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_Throw(ctx: *mut JSContext, obj: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetException(ctx: *mut JSContext) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_HasException(ctx: *mut JSContext) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsError(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsUncatchableError(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_SetUncatchableError(ctx: *mut JSContext, val: JSValue);
}
unsafe extern "C" {
    pub fn JS_ClearUncatchableError(ctx: *mut JSContext, val: JSValue);
}
unsafe extern "C" {
    pub fn JS_ResetUncatchableError(ctx: *mut JSContext);
}
unsafe extern "C" {
    pub fn JS_NewError(ctx: *mut JSContext) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewInternalError(
        ctx: *mut JSContext,
        fmt: *const ::core::ffi::c_char,
        ...
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewPlainError(ctx: *mut JSContext, fmt: *const ::core::ffi::c_char, ...) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewRangeError(ctx: *mut JSContext, fmt: *const ::core::ffi::c_char, ...) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewReferenceError(
        ctx: *mut JSContext,
        fmt: *const ::core::ffi::c_char,
        ...
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewSyntaxError(ctx: *mut JSContext, fmt: *const ::core::ffi::c_char, ...) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewTypeError(ctx: *mut JSContext, fmt: *const ::core::ffi::c_char, ...) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ThrowInternalError(
        ctx: *mut JSContext,
        fmt: *const ::core::ffi::c_char,
        ...
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ThrowPlainError(ctx: *mut JSContext, fmt: *const ::core::ffi::c_char, ...)
    -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ThrowRangeError(ctx: *mut JSContext, fmt: *const ::core::ffi::c_char, ...)
    -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ThrowReferenceError(
        ctx: *mut JSContext,
        fmt: *const ::core::ffi::c_char,
        ...
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ThrowSyntaxError(
        ctx: *mut JSContext,
        fmt: *const ::core::ffi::c_char,
        ...
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ThrowTypeError(ctx: *mut JSContext, fmt: *const ::core::ffi::c_char, ...) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ThrowDOMException(
        ctx: *mut JSContext,
        name: *const ::core::ffi::c_char,
        fmt: *const ::core::ffi::c_char,
        ...
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ThrowOutOfMemory(ctx: *mut JSContext) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_FreeValue(ctx: *mut JSContext, v: JSValue);
}
unsafe extern "C" {
    pub fn JS_FreeValueRT(rt: *mut JSRuntime, v: JSValue);
}
unsafe extern "C" {
    pub fn JS_DupValue(ctx: *mut JSContext, v: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_DupValueRT(rt: *mut JSRuntime, v: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ToBool(ctx: *mut JSContext, val: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_ToNumber(ctx: *mut JSContext, val: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ToInt32(ctx: *mut JSContext, pres: *mut i32, val: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_ToInt64(ctx: *mut JSContext, pres: *mut i64, val: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_ToIndex(ctx: *mut JSContext, plen: *mut u64, val: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_ToFloat64(ctx: *mut JSContext, pres: *mut f64, val: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_ToBigInt64(ctx: *mut JSContext, pres: *mut i64, val: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_ToBigUint64(ctx: *mut JSContext, pres: *mut u64, val: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_ToInt64Ext(ctx: *mut JSContext, pres: *mut i64, val: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_NewStringLen(
        ctx: *mut JSContext,
        str1: *const ::core::ffi::c_char,
        len1: size_t,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewTwoByteString(ctx: *mut JSContext, buf: *const u16, len: size_t) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewAtomString(ctx: *mut JSContext, str_: *const ::core::ffi::c_char) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ToString(ctx: *mut JSContext, val: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ToPropertyKey(ctx: *mut JSContext, val: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ToCStringLen2(
        ctx: *mut JSContext,
        plen: *mut size_t,
        val1: JSValue,
        cesu8: bool,
    ) -> *const ::core::ffi::c_char;
}
unsafe extern "C" {
    pub fn JS_FreeCString(ctx: *mut JSContext, ptr: *const ::core::ffi::c_char);
}
unsafe extern "C" {
    pub fn JS_NewObjectProtoClass(
        ctx: *mut JSContext,
        proto: JSValue,
        class_id: JSClassID,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewObjectClass(ctx: *mut JSContext, class_id: JSClassID) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewObjectProto(ctx: *mut JSContext, proto: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewObject(ctx: *mut JSContext) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewObjectFrom(
        ctx: *mut JSContext,
        count: ::core::ffi::c_int,
        props: *const JSAtom,
        values: *const JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewObjectFromStr(
        ctx: *mut JSContext,
        count: ::core::ffi::c_int,
        props: *mut *const ::core::ffi::c_char,
        values: *const JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ToObject(ctx: *mut JSContext, val: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ToObjectString(ctx: *mut JSContext, val: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_IsFunction(ctx: *mut JSContext, val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsConstructor(ctx: *mut JSContext, val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_SetConstructorBit(ctx: *mut JSContext, func_obj: JSValue, val: bool) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsRegExp(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsMap(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsSet(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsWeakRef(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsWeakSet(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsWeakMap(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsDataView(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_NewArray(ctx: *mut JSContext) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewArrayFrom(
        ctx: *mut JSContext,
        count: ::core::ffi::c_int,
        values: *const JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_IsArray(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_IsProxy(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_GetProxyTarget(ctx: *mut JSContext, proxy: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetProxyHandler(ctx: *mut JSContext, proxy: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewDate(ctx: *mut JSContext, epoch_ms: f64) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_IsDate(v: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_GetProperty(ctx: *mut JSContext, this_obj: JSValue, prop: JSAtom) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetPropertyUint32(ctx: *mut JSContext, this_obj: JSValue, idx: u32) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetPropertyInt64(ctx: *mut JSContext, this_obj: JSValue, idx: i64) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetPropertyStr(
        ctx: *mut JSContext,
        this_obj: JSValue,
        prop: *const ::core::ffi::c_char,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_SetProperty(
        ctx: *mut JSContext,
        this_obj: JSValue,
        prop: JSAtom,
        val: JSValue,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_SetPropertyUint32(
        ctx: *mut JSContext,
        this_obj: JSValue,
        idx: u32,
        val: JSValue,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_SetPropertyInt64(
        ctx: *mut JSContext,
        this_obj: JSValue,
        idx: i64,
        val: JSValue,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_SetPropertyStr(
        ctx: *mut JSContext,
        this_obj: JSValue,
        prop: *const ::core::ffi::c_char,
        val: JSValue,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_HasProperty(
        ctx: *mut JSContext,
        this_obj: JSValue,
        prop: JSAtom,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_IsExtensible(ctx: *mut JSContext, obj: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_PreventExtensions(ctx: *mut JSContext, obj: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_DeleteProperty(
        ctx: *mut JSContext,
        obj: JSValue,
        prop: JSAtom,
        flags: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_SetPrototype(
        ctx: *mut JSContext,
        obj: JSValue,
        proto_val: JSValue,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_GetPrototype(ctx: *mut JSContext, val: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetLength(ctx: *mut JSContext, obj: JSValue, pres: *mut i64) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_SetLength(ctx: *mut JSContext, obj: JSValue, len: i64) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_SealObject(ctx: *mut JSContext, obj: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_FreezeObject(ctx: *mut JSContext, obj: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_GetOwnPropertyNames(
        ctx: *mut JSContext,
        ptab: *mut *mut JSPropertyEnum,
        plen: *mut u32,
        obj: JSValue,
        flags: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_GetOwnProperty(
        ctx: *mut JSContext,
        desc: *mut JSPropertyDescriptor,
        obj: JSValue,
        prop: JSAtom,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_FreePropertyEnum(ctx: *mut JSContext, tab: *mut JSPropertyEnum, len: u32);
}
unsafe extern "C" {
    pub fn JS_Call(
        ctx: *mut JSContext,
        func_obj: JSValue,
        this_obj: JSValue,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_Invoke(
        ctx: *mut JSContext,
        this_val: JSValue,
        atom: JSAtom,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_CallConstructor(
        ctx: *mut JSContext,
        func_obj: JSValue,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_CallConstructor2(
        ctx: *mut JSContext,
        func_obj: JSValue,
        new_target: JSValue,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_DetectModule(input: *const ::core::ffi::c_char, input_len: size_t) -> bool;
}
unsafe extern "C" {
    pub fn JS_Eval(
        ctx: *mut JSContext,
        input: *const ::core::ffi::c_char,
        input_len: size_t,
        filename: *const ::core::ffi::c_char,
        eval_flags: ::core::ffi::c_int,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_Eval2(
        ctx: *mut JSContext,
        input: *const ::core::ffi::c_char,
        input_len: size_t,
        options: *mut JSEvalOptions,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_EvalThis(
        ctx: *mut JSContext,
        this_obj: JSValue,
        input: *const ::core::ffi::c_char,
        input_len: size_t,
        filename: *const ::core::ffi::c_char,
        eval_flags: ::core::ffi::c_int,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_EvalThis2(
        ctx: *mut JSContext,
        this_obj: JSValue,
        input: *const ::core::ffi::c_char,
        input_len: size_t,
        options: *mut JSEvalOptions,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetGlobalObject(ctx: *mut JSContext) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_IsInstanceOf(ctx: *mut JSContext, val: JSValue, obj: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_DefineProperty(
        ctx: *mut JSContext,
        this_obj: JSValue,
        prop: JSAtom,
        val: JSValue,
        getter: JSValue,
        setter: JSValue,
        flags: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_DefinePropertyValue(
        ctx: *mut JSContext,
        this_obj: JSValue,
        prop: JSAtom,
        val: JSValue,
        flags: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_DefinePropertyValueUint32(
        ctx: *mut JSContext,
        this_obj: JSValue,
        idx: u32,
        val: JSValue,
        flags: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_DefinePropertyValueStr(
        ctx: *mut JSContext,
        this_obj: JSValue,
        prop: *const ::core::ffi::c_char,
        val: JSValue,
        flags: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_DefinePropertyGetSet(
        ctx: *mut JSContext,
        this_obj: JSValue,
        prop: JSAtom,
        getter: JSValue,
        setter: JSValue,
        flags: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_SetOpaque(obj: JSValue, opaque: *mut ::core::ffi::c_void) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_GetOpaque(obj: JSValue, class_id: JSClassID) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn JS_GetOpaque2(
        ctx: *mut JSContext,
        obj: JSValue,
        class_id: JSClassID,
    ) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn JS_GetAnyOpaque(obj: JSValue, class_id: *mut JSClassID) -> *mut ::core::ffi::c_void;
}
unsafe extern "C" {
    pub fn JS_ParseJSON(
        ctx: *mut JSContext,
        buf: *const ::core::ffi::c_char,
        buf_len: size_t,
        filename: *const ::core::ffi::c_char,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_JSONStringify(
        ctx: *mut JSContext,
        obj: JSValue,
        replacer: JSValue,
        space0: JSValue,
    ) -> JSValue;
}
pub type JSFreeArrayBufferDataFunc = ::core::option::Option<
    unsafe extern "C" fn(
        rt: *mut JSRuntime,
        opaque: *mut ::core::ffi::c_void,
        ptr: *mut ::core::ffi::c_void,
    ),
>;
unsafe extern "C" {
    pub fn JS_NewArrayBuffer(
        ctx: *mut JSContext,
        buf: *mut u8,
        len: size_t,
        free_func: JSFreeArrayBufferDataFunc,
        opaque: *mut ::core::ffi::c_void,
        is_shared: bool,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewArrayBufferCopy(ctx: *mut JSContext, buf: *const u8, len: size_t) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_DetachArrayBuffer(ctx: *mut JSContext, obj: JSValue);
}
unsafe extern "C" {
    pub fn JS_GetArrayBuffer(ctx: *mut JSContext, psize: *mut size_t, obj: JSValue) -> *mut u8;
}
unsafe extern "C" {
    pub fn JS_IsArrayBuffer(obj: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_GetUint8Array(ctx: *mut JSContext, psize: *mut size_t, obj: JSValue) -> *mut u8;
}
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_UINT8C: JSTypedArrayEnum = 0;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_INT8: JSTypedArrayEnum = 1;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_UINT8: JSTypedArrayEnum = 2;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_INT16: JSTypedArrayEnum = 3;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_UINT16: JSTypedArrayEnum = 4;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_INT32: JSTypedArrayEnum = 5;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_UINT32: JSTypedArrayEnum = 6;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_BIG_INT64: JSTypedArrayEnum = 7;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_BIG_UINT64: JSTypedArrayEnum = 8;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_FLOAT16: JSTypedArrayEnum = 9;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_FLOAT32: JSTypedArrayEnum = 10;
pub const JSTypedArrayEnum_JS_TYPED_ARRAY_FLOAT64: JSTypedArrayEnum = 11;
pub type JSTypedArrayEnum = ::core::ffi::c_uint;
unsafe extern "C" {
    pub fn JS_NewTypedArray(
        ctx: *mut JSContext,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
        array_type: JSTypedArrayEnum,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetTypedArrayBuffer(
        ctx: *mut JSContext,
        obj: JSValue,
        pbyte_offset: *mut size_t,
        pbyte_length: *mut size_t,
        pbytes_per_element: *mut size_t,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewUint8Array(
        ctx: *mut JSContext,
        buf: *mut u8,
        len: size_t,
        free_func: JSFreeArrayBufferDataFunc,
        opaque: *mut ::core::ffi::c_void,
        is_shared: bool,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetTypedArrayType(obj: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_NewUint8ArrayCopy(ctx: *mut JSContext, buf: *const u8, len: size_t) -> JSValue;
}
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSSharedArrayBufferFunctions {
    pub sab_alloc: ::core::option::Option<
        unsafe extern "C" fn(
            opaque: *mut ::core::ffi::c_void,
            size: size_t,
        ) -> *mut ::core::ffi::c_void,
    >,
    pub sab_free: ::core::option::Option<
        unsafe extern "C" fn(opaque: *mut ::core::ffi::c_void, ptr: *mut ::core::ffi::c_void),
    >,
    pub sab_dup: ::core::option::Option<
        unsafe extern "C" fn(opaque: *mut ::core::ffi::c_void, ptr: *mut ::core::ffi::c_void),
    >,
    pub sab_opaque: *mut ::core::ffi::c_void,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSSharedArrayBufferFunctions"]
        [::core::mem::size_of::<JSSharedArrayBufferFunctions>() - 32usize];
    ["Alignment of JSSharedArrayBufferFunctions"]
        [::core::mem::align_of::<JSSharedArrayBufferFunctions>() - 8usize];
    ["Offset of field: JSSharedArrayBufferFunctions::sab_alloc"]
        [::core::mem::offset_of!(JSSharedArrayBufferFunctions, sab_alloc) - 0usize];
    ["Offset of field: JSSharedArrayBufferFunctions::sab_free"]
        [::core::mem::offset_of!(JSSharedArrayBufferFunctions, sab_free) - 8usize];
    ["Offset of field: JSSharedArrayBufferFunctions::sab_dup"]
        [::core::mem::offset_of!(JSSharedArrayBufferFunctions, sab_dup) - 16usize];
    ["Offset of field: JSSharedArrayBufferFunctions::sab_opaque"]
        [::core::mem::offset_of!(JSSharedArrayBufferFunctions, sab_opaque) - 24usize];
};
unsafe extern "C" {
    pub fn JS_SetSharedArrayBufferFunctions(
        rt: *mut JSRuntime,
        sf: *const JSSharedArrayBufferFunctions,
    );
}
pub const JSPromiseStateEnum_JS_PROMISE_PENDING: JSPromiseStateEnum = 0;
pub const JSPromiseStateEnum_JS_PROMISE_FULFILLED: JSPromiseStateEnum = 1;
pub const JSPromiseStateEnum_JS_PROMISE_REJECTED: JSPromiseStateEnum = 2;
pub type JSPromiseStateEnum = ::core::ffi::c_uint;
unsafe extern "C" {
    pub fn JS_NewPromiseCapability(ctx: *mut JSContext, resolving_funcs: *mut JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_PromiseState(ctx: *mut JSContext, promise: JSValue) -> JSPromiseStateEnum;
}
unsafe extern "C" {
    pub fn JS_PromiseResult(ctx: *mut JSContext, promise: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_IsPromise(val: JSValue) -> bool;
}
unsafe extern "C" {
    pub fn JS_NewSymbol(
        ctx: *mut JSContext,
        description: *const ::core::ffi::c_char,
        is_global: bool,
    ) -> JSValue;
}
pub const JSPromiseHookType_JS_PROMISE_HOOK_INIT: JSPromiseHookType = 0;
pub const JSPromiseHookType_JS_PROMISE_HOOK_BEFORE: JSPromiseHookType = 1;
pub const JSPromiseHookType_JS_PROMISE_HOOK_AFTER: JSPromiseHookType = 2;
pub const JSPromiseHookType_JS_PROMISE_HOOK_RESOLVE: JSPromiseHookType = 3;
pub type JSPromiseHookType = ::core::ffi::c_uint;
pub type JSPromiseHook = ::core::option::Option<
    unsafe extern "C" fn(
        ctx: *mut JSContext,
        type_: JSPromiseHookType,
        promise: JSValue,
        parent_promise: JSValue,
        opaque: *mut ::core::ffi::c_void,
    ),
>;
unsafe extern "C" {
    pub fn JS_SetPromiseHook(
        rt: *mut JSRuntime,
        promise_hook: JSPromiseHook,
        opaque: *mut ::core::ffi::c_void,
    );
}
pub type JSHostPromiseRejectionTracker = ::core::option::Option<
    unsafe extern "C" fn(
        ctx: *mut JSContext,
        promise: JSValue,
        reason: JSValue,
        is_handled: bool,
        opaque: *mut ::core::ffi::c_void,
    ),
>;
unsafe extern "C" {
    pub fn JS_SetHostPromiseRejectionTracker(
        rt: *mut JSRuntime,
        cb: JSHostPromiseRejectionTracker,
        opaque: *mut ::core::ffi::c_void,
    );
}
pub type JSInterruptHandler = ::core::option::Option<
    unsafe extern "C" fn(
        rt: *mut JSRuntime,
        opaque: *mut ::core::ffi::c_void,
    ) -> ::core::ffi::c_int,
>;
unsafe extern "C" {
    pub fn JS_SetInterruptHandler(
        rt: *mut JSRuntime,
        cb: JSInterruptHandler,
        opaque: *mut ::core::ffi::c_void,
    );
}
unsafe extern "C" {
    pub fn JS_SetCanBlock(rt: *mut JSRuntime, can_block: bool);
}
unsafe extern "C" {
    pub fn JS_SetIsHTMLDDA(ctx: *mut JSContext, obj: JSValue);
}
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSModuleDef {
    _unused: [u8; 0],
}
pub type JSModuleNormalizeFunc = ::core::option::Option<
    unsafe extern "C" fn(
        ctx: *mut JSContext,
        module_base_name: *const ::core::ffi::c_char,
        module_name: *const ::core::ffi::c_char,
        opaque: *mut ::core::ffi::c_void,
    ) -> *mut ::core::ffi::c_char,
>;
pub type JSModuleLoaderFunc = ::core::option::Option<
    unsafe extern "C" fn(
        ctx: *mut JSContext,
        module_name: *const ::core::ffi::c_char,
        opaque: *mut ::core::ffi::c_void,
    ) -> *mut JSModuleDef,
>;
unsafe extern "C" {
    pub fn JS_SetModuleLoaderFunc(
        rt: *mut JSRuntime,
        module_normalize: JSModuleNormalizeFunc,
        module_loader: JSModuleLoaderFunc,
        opaque: *mut ::core::ffi::c_void,
    );
}
unsafe extern "C" {
    pub fn JS_GetImportMeta(ctx: *mut JSContext, m: *mut JSModuleDef) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_GetModuleName(ctx: *mut JSContext, m: *mut JSModuleDef) -> JSAtom;
}
unsafe extern "C" {
    pub fn JS_GetModuleNamespace(ctx: *mut JSContext, m: *mut JSModuleDef) -> JSValue;
}
pub type JSJobFunc = ::core::option::Option<
    unsafe extern "C" fn(
        ctx: *mut JSContext,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
    ) -> JSValue,
>;
unsafe extern "C" {
    pub fn JS_EnqueueJob(
        ctx: *mut JSContext,
        job_func: JSJobFunc,
        argc: ::core::ffi::c_int,
        argv: *mut JSValue,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_IsJobPending(rt: *mut JSRuntime) -> bool;
}
unsafe extern "C" {
    pub fn JS_ExecutePendingJob(
        rt: *mut JSRuntime,
        pctx: *mut *mut JSContext,
    ) -> ::core::ffi::c_int;
}
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSSABTab {
    pub tab: *mut *mut u8,
    pub len: size_t,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSSABTab"][::core::mem::size_of::<JSSABTab>() - 16usize];
    ["Alignment of JSSABTab"][::core::mem::align_of::<JSSABTab>() - 8usize];
    ["Offset of field: JSSABTab::tab"][::core::mem::offset_of!(JSSABTab, tab) - 0usize];
    ["Offset of field: JSSABTab::len"][::core::mem::offset_of!(JSSABTab, len) - 8usize];
};
unsafe extern "C" {
    pub fn JS_WriteObject(
        ctx: *mut JSContext,
        psize: *mut size_t,
        obj: JSValue,
        flags: ::core::ffi::c_int,
    ) -> *mut u8;
}
unsafe extern "C" {
    pub fn JS_WriteObject2(
        ctx: *mut JSContext,
        psize: *mut size_t,
        obj: JSValue,
        flags: ::core::ffi::c_int,
        psab_tab: *mut JSSABTab,
    ) -> *mut u8;
}
unsafe extern "C" {
    pub fn JS_ReadObject(
        ctx: *mut JSContext,
        buf: *const u8,
        buf_len: size_t,
        flags: ::core::ffi::c_int,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ReadObject2(
        ctx: *mut JSContext,
        buf: *const u8,
        buf_len: size_t,
        flags: ::core::ffi::c_int,
        psab_tab: *mut JSSABTab,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_EvalFunction(ctx: *mut JSContext, fun_obj: JSValue) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_ResolveModule(ctx: *mut JSContext, obj: JSValue) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_GetScriptOrModuleName(
        ctx: *mut JSContext,
        n_stack_levels: ::core::ffi::c_int,
    ) -> JSAtom;
}
unsafe extern "C" {
    pub fn JS_LoadModule(
        ctx: *mut JSContext,
        basename: *const ::core::ffi::c_char,
        filename: *const ::core::ffi::c_char,
    ) -> JSValue;
}
pub const JSCFunctionEnum_JS_CFUNC_generic: JSCFunctionEnum = 0;
pub const JSCFunctionEnum_JS_CFUNC_generic_magic: JSCFunctionEnum = 1;
pub const JSCFunctionEnum_JS_CFUNC_constructor: JSCFunctionEnum = 2;
pub const JSCFunctionEnum_JS_CFUNC_constructor_magic: JSCFunctionEnum = 3;
pub const JSCFunctionEnum_JS_CFUNC_constructor_or_func: JSCFunctionEnum = 4;
pub const JSCFunctionEnum_JS_CFUNC_constructor_or_func_magic: JSCFunctionEnum = 5;
pub const JSCFunctionEnum_JS_CFUNC_f_f: JSCFunctionEnum = 6;
pub const JSCFunctionEnum_JS_CFUNC_f_f_f: JSCFunctionEnum = 7;
pub const JSCFunctionEnum_JS_CFUNC_getter: JSCFunctionEnum = 8;
pub const JSCFunctionEnum_JS_CFUNC_setter: JSCFunctionEnum = 9;
pub const JSCFunctionEnum_JS_CFUNC_getter_magic: JSCFunctionEnum = 10;
pub const JSCFunctionEnum_JS_CFUNC_setter_magic: JSCFunctionEnum = 11;
pub const JSCFunctionEnum_JS_CFUNC_iterator_next: JSCFunctionEnum = 12;
pub type JSCFunctionEnum = ::core::ffi::c_uint;
#[repr(C)]
#[derive(Copy, Clone)]
pub union JSCFunctionType {
    pub generic: JSCFunction,
    pub generic_magic: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            this_val: JSValue,
            argc: ::core::ffi::c_int,
            argv: *mut JSValue,
            magic: ::core::ffi::c_int,
        ) -> JSValue,
    >,
    pub constructor: JSCFunction,
    pub constructor_magic: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            new_target: JSValue,
            argc: ::core::ffi::c_int,
            argv: *mut JSValue,
            magic: ::core::ffi::c_int,
        ) -> JSValue,
    >,
    pub constructor_or_func: JSCFunction,
    pub f_f: ::core::option::Option<unsafe extern "C" fn(arg1: f64) -> f64>,
    pub f_f_f: ::core::option::Option<unsafe extern "C" fn(arg1: f64, arg2: f64) -> f64>,
    pub getter: ::core::option::Option<
        unsafe extern "C" fn(ctx: *mut JSContext, this_val: JSValue) -> JSValue,
    >,
    pub setter: ::core::option::Option<
        unsafe extern "C" fn(ctx: *mut JSContext, this_val: JSValue, val: JSValue) -> JSValue,
    >,
    pub getter_magic: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            this_val: JSValue,
            magic: ::core::ffi::c_int,
        ) -> JSValue,
    >,
    pub setter_magic: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            this_val: JSValue,
            val: JSValue,
            magic: ::core::ffi::c_int,
        ) -> JSValue,
    >,
    pub iterator_next: ::core::option::Option<
        unsafe extern "C" fn(
            ctx: *mut JSContext,
            this_val: JSValue,
            argc: ::core::ffi::c_int,
            argv: *mut JSValue,
            pdone: *mut ::core::ffi::c_int,
            magic: ::core::ffi::c_int,
        ) -> JSValue,
    >,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSCFunctionType"][::core::mem::size_of::<JSCFunctionType>() - 8usize];
    ["Alignment of JSCFunctionType"][::core::mem::align_of::<JSCFunctionType>() - 8usize];
    ["Offset of field: JSCFunctionType::generic"]
        [::core::mem::offset_of!(JSCFunctionType, generic) - 0usize];
    ["Offset of field: JSCFunctionType::generic_magic"]
        [::core::mem::offset_of!(JSCFunctionType, generic_magic) - 0usize];
    ["Offset of field: JSCFunctionType::constructor"]
        [::core::mem::offset_of!(JSCFunctionType, constructor) - 0usize];
    ["Offset of field: JSCFunctionType::constructor_magic"]
        [::core::mem::offset_of!(JSCFunctionType, constructor_magic) - 0usize];
    ["Offset of field: JSCFunctionType::constructor_or_func"]
        [::core::mem::offset_of!(JSCFunctionType, constructor_or_func) - 0usize];
    ["Offset of field: JSCFunctionType::f_f"]
        [::core::mem::offset_of!(JSCFunctionType, f_f) - 0usize];
    ["Offset of field: JSCFunctionType::f_f_f"]
        [::core::mem::offset_of!(JSCFunctionType, f_f_f) - 0usize];
    ["Offset of field: JSCFunctionType::getter"]
        [::core::mem::offset_of!(JSCFunctionType, getter) - 0usize];
    ["Offset of field: JSCFunctionType::setter"]
        [::core::mem::offset_of!(JSCFunctionType, setter) - 0usize];
    ["Offset of field: JSCFunctionType::getter_magic"]
        [::core::mem::offset_of!(JSCFunctionType, getter_magic) - 0usize];
    ["Offset of field: JSCFunctionType::setter_magic"]
        [::core::mem::offset_of!(JSCFunctionType, setter_magic) - 0usize];
    ["Offset of field: JSCFunctionType::iterator_next"]
        [::core::mem::offset_of!(JSCFunctionType, iterator_next) - 0usize];
};
unsafe extern "C" {
    pub fn JS_NewCFunction2(
        ctx: *mut JSContext,
        func: JSCFunction,
        name: *const ::core::ffi::c_char,
        length: ::core::ffi::c_int,
        cproto: JSCFunctionEnum,
        magic: ::core::ffi::c_int,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewCFunction3(
        ctx: *mut JSContext,
        func: JSCFunction,
        name: *const ::core::ffi::c_char,
        length: ::core::ffi::c_int,
        cproto: JSCFunctionEnum,
        magic: ::core::ffi::c_int,
        proto_val: JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewCFunctionData(
        ctx: *mut JSContext,
        func: JSCFunctionData,
        length: ::core::ffi::c_int,
        magic: ::core::ffi::c_int,
        data_len: ::core::ffi::c_int,
        data: *mut JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_NewCFunctionData2(
        ctx: *mut JSContext,
        func: JSCFunctionData,
        name: *const ::core::ffi::c_char,
        length: ::core::ffi::c_int,
        magic: ::core::ffi::c_int,
        data_len: ::core::ffi::c_int,
        data: *mut JSValue,
    ) -> JSValue;
}
unsafe extern "C" {
    pub fn JS_SetConstructor(ctx: *mut JSContext, func_obj: JSValue, proto: JSValue);
}
#[repr(C)]
#[derive(Copy, Clone)]
pub struct JSCFunctionListEntry {
    pub name: *const ::core::ffi::c_char,
    pub prop_flags: u8,
    pub def_type: u8,
    pub magic: i16,
    pub u: JSCFunctionListEntry__bindgen_ty_1,
}
#[repr(C)]
#[derive(Copy, Clone)]
pub union JSCFunctionListEntry__bindgen_ty_1 {
    pub func: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1,
    pub getset: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2,
    pub alias: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3,
    pub prop_list: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4,
    pub str_: *const ::core::ffi::c_char,
    pub i32_: i32,
    pub i64_: i64,
    pub u64_: u64,
    pub f64_: f64,
}
#[repr(C)]
#[derive(Copy, Clone)]
pub struct JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1 {
    pub length: u8,
    pub cproto: u8,
    pub cfunc: JSCFunctionType,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1"]
        [::core::mem::size_of::<JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1>() - 16usize];
    ["Alignment of JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1"]
        [::core::mem::align_of::<JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1>() - 8usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1::length"][::core::mem::offset_of!(
        JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1,
        length
    ) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1::cproto"][::core::mem::offset_of!(
        JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1,
        cproto
    ) - 1usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1::cfunc"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_1, cfunc) - 8usize];
};
#[repr(C)]
#[derive(Copy, Clone)]
pub struct JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2 {
    pub get: JSCFunctionType,
    pub set: JSCFunctionType,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2"]
        [::core::mem::size_of::<JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2>() - 16usize];
    ["Alignment of JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2"]
        [::core::mem::align_of::<JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2>() - 8usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2::get"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2, get) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2::set"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_2, set) - 8usize];
};
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3 {
    pub name: *const ::core::ffi::c_char,
    pub base: ::core::ffi::c_int,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3"]
        [::core::mem::size_of::<JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3>() - 16usize];
    ["Alignment of JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3"]
        [::core::mem::align_of::<JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3>() - 8usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3::name"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3, name) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3::base"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_3, base) - 8usize];
};
#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4 {
    pub tab: *const JSCFunctionListEntry,
    pub len: ::core::ffi::c_int,
}
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4"]
        [::core::mem::size_of::<JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4>() - 16usize];
    ["Alignment of JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4"]
        [::core::mem::align_of::<JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4>() - 8usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4::tab"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4, tab) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4::len"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1__bindgen_ty_4, len) - 8usize];
};
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSCFunctionListEntry__bindgen_ty_1"]
        [::core::mem::size_of::<JSCFunctionListEntry__bindgen_ty_1>() - 16usize];
    ["Alignment of JSCFunctionListEntry__bindgen_ty_1"]
        [::core::mem::align_of::<JSCFunctionListEntry__bindgen_ty_1>() - 8usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1::func"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1, func) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1::getset"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1, getset) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1::alias"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1, alias) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1::prop_list"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1, prop_list) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1::str_"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1, str_) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1::i32_"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1, i32_) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1::i64_"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1, i64_) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1::u64_"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1, u64_) - 0usize];
    ["Offset of field: JSCFunctionListEntry__bindgen_ty_1::f64_"]
        [::core::mem::offset_of!(JSCFunctionListEntry__bindgen_ty_1, f64_) - 0usize];
};
#[allow(clippy::unnecessary_operation, clippy::identity_op)]
const _: () = {
    ["Size of JSCFunctionListEntry"][::core::mem::size_of::<JSCFunctionListEntry>() - 32usize];
    ["Alignment of JSCFunctionListEntry"][::core::mem::align_of::<JSCFunctionListEntry>() - 8usize];
    ["Offset of field: JSCFunctionListEntry::name"]
        [::core::mem::offset_of!(JSCFunctionListEntry, name) - 0usize];
    ["Offset of field: JSCFunctionListEntry::prop_flags"]
        [::core::mem::offset_of!(JSCFunctionListEntry, prop_flags) - 8usize];
    ["Offset of field: JSCFunctionListEntry::def_type"]
        [::core::mem::offset_of!(JSCFunctionListEntry, def_type) - 9usize];
    ["Offset of field: JSCFunctionListEntry::magic"]
        [::core::mem::offset_of!(JSCFunctionListEntry, magic) - 10usize];
    ["Offset of field: JSCFunctionListEntry::u"]
        [::core::mem::offset_of!(JSCFunctionListEntry, u) - 16usize];
};
unsafe extern "C" {
    pub fn JS_SetPropertyFunctionList(
        ctx: *mut JSContext,
        obj: JSValue,
        tab: *const JSCFunctionListEntry,
        len: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
pub type JSModuleInitFunc = ::core::option::Option<
    unsafe extern "C" fn(ctx: *mut JSContext, m: *mut JSModuleDef) -> ::core::ffi::c_int,
>;
unsafe extern "C" {
    pub fn JS_NewCModule(
        ctx: *mut JSContext,
        name_str: *const ::core::ffi::c_char,
        func: JSModuleInitFunc,
    ) -> *mut JSModuleDef;
}
unsafe extern "C" {
    pub fn JS_AddModuleExport(
        ctx: *mut JSContext,
        m: *mut JSModuleDef,
        name_str: *const ::core::ffi::c_char,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_AddModuleExportList(
        ctx: *mut JSContext,
        m: *mut JSModuleDef,
        tab: *const JSCFunctionListEntry,
        len: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_SetModuleExport(
        ctx: *mut JSContext,
        m: *mut JSModuleDef,
        export_name: *const ::core::ffi::c_char,
        val: JSValue,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_SetModuleExportList(
        ctx: *mut JSContext,
        m: *mut JSModuleDef,
        tab: *const JSCFunctionListEntry,
        len: ::core::ffi::c_int,
    ) -> ::core::ffi::c_int;
}
unsafe extern "C" {
    pub fn JS_GetVersion() -> *const ::core::ffi::c_char;
}
unsafe extern "C" {
    pub fn js_std_cmd(cmd: ::core::ffi::c_int, ...) -> usize;
}
pub const __JS_ATOM_NULL: _bindgen_ty_2 = 0;
pub const JS_ATOM_null: _bindgen_ty_2 = 1;
pub const JS_ATOM_false: _bindgen_ty_2 = 2;
pub const JS_ATOM_true: _bindgen_ty_2 = 3;
pub const JS_ATOM_if: _bindgen_ty_2 = 4;
pub const JS_ATOM_else: _bindgen_ty_2 = 5;
pub const JS_ATOM_return: _bindgen_ty_2 = 6;
pub const JS_ATOM_var: _bindgen_ty_2 = 7;
pub const JS_ATOM_this: _bindgen_ty_2 = 8;
pub const JS_ATOM_delete: _bindgen_ty_2 = 9;
pub const JS_ATOM_void: _bindgen_ty_2 = 10;
pub const JS_ATOM_typeof: _bindgen_ty_2 = 11;
pub const JS_ATOM_new: _bindgen_ty_2 = 12;
pub const JS_ATOM_in: _bindgen_ty_2 = 13;
pub const JS_ATOM_instanceof: _bindgen_ty_2 = 14;
pub const JS_ATOM_do: _bindgen_ty_2 = 15;
pub const JS_ATOM_while: _bindgen_ty_2 = 16;
pub const JS_ATOM_for: _bindgen_ty_2 = 17;
pub const JS_ATOM_break: _bindgen_ty_2 = 18;
pub const JS_ATOM_continue: _bindgen_ty_2 = 19;
pub const JS_ATOM_switch: _bindgen_ty_2 = 20;
pub const JS_ATOM_case: _bindgen_ty_2 = 21;
pub const JS_ATOM_default: _bindgen_ty_2 = 22;
pub const JS_ATOM_throw: _bindgen_ty_2 = 23;
pub const JS_ATOM_try: _bindgen_ty_2 = 24;
pub const JS_ATOM_catch: _bindgen_ty_2 = 25;
pub const JS_ATOM_finally: _bindgen_ty_2 = 26;
pub const JS_ATOM_function: _bindgen_ty_2 = 27;
pub const JS_ATOM_debugger: _bindgen_ty_2 = 28;
pub const JS_ATOM_with: _bindgen_ty_2 = 29;
pub const JS_ATOM_class: _bindgen_ty_2 = 30;
pub const JS_ATOM_const: _bindgen_ty_2 = 31;
pub const JS_ATOM_enum: _bindgen_ty_2 = 32;
pub const JS_ATOM_export: _bindgen_ty_2 = 33;
pub const JS_ATOM_extends: _bindgen_ty_2 = 34;
pub const JS_ATOM_import: _bindgen_ty_2 = 35;
pub const JS_ATOM_super: _bindgen_ty_2 = 36;
pub const JS_ATOM_implements: _bindgen_ty_2 = 37;
pub const JS_ATOM_interface: _bindgen_ty_2 = 38;
pub const JS_ATOM_let: _bindgen_ty_2 = 39;
pub const JS_ATOM_package: _bindgen_ty_2 = 40;
pub const JS_ATOM_private: _bindgen_ty_2 = 41;
pub const JS_ATOM_protected: _bindgen_ty_2 = 42;
pub const JS_ATOM_public: _bindgen_ty_2 = 43;
pub const JS_ATOM_static: _bindgen_ty_2 = 44;
pub const JS_ATOM_yield: _bindgen_ty_2 = 45;
pub const JS_ATOM_await: _bindgen_ty_2 = 46;
pub const JS_ATOM_empty_string: _bindgen_ty_2 = 47;
pub const JS_ATOM_keys: _bindgen_ty_2 = 48;
pub const JS_ATOM_size: _bindgen_ty_2 = 49;
pub const JS_ATOM_length: _bindgen_ty_2 = 50;
pub const JS_ATOM_message: _bindgen_ty_2 = 51;
pub const JS_ATOM_cause: _bindgen_ty_2 = 52;
pub const JS_ATOM_errors: _bindgen_ty_2 = 53;
pub const JS_ATOM_stack: _bindgen_ty_2 = 54;
pub const JS_ATOM_name: _bindgen_ty_2 = 55;
pub const JS_ATOM_toString: _bindgen_ty_2 = 56;
pub const JS_ATOM_toLocaleString: _bindgen_ty_2 = 57;
pub const JS_ATOM_valueOf: _bindgen_ty_2 = 58;
pub const JS_ATOM_eval: _bindgen_ty_2 = 59;
pub const JS_ATOM_prototype: _bindgen_ty_2 = 60;
pub const JS_ATOM_constructor: _bindgen_ty_2 = 61;
pub const JS_ATOM_configurable: _bindgen_ty_2 = 62;
pub const JS_ATOM_writable: _bindgen_ty_2 = 63;
pub const JS_ATOM_enumerable: _bindgen_ty_2 = 64;
pub const JS_ATOM_value: _bindgen_ty_2 = 65;
pub const JS_ATOM_get: _bindgen_ty_2 = 66;
pub const JS_ATOM_set: _bindgen_ty_2 = 67;
pub const JS_ATOM_of: _bindgen_ty_2 = 68;
pub const JS_ATOM___proto__: _bindgen_ty_2 = 69;
pub const JS_ATOM_undefined: _bindgen_ty_2 = 70;
pub const JS_ATOM_number: _bindgen_ty_2 = 71;
pub const JS_ATOM_boolean: _bindgen_ty_2 = 72;
pub const JS_ATOM_string: _bindgen_ty_2 = 73;
pub const JS_ATOM_object: _bindgen_ty_2 = 74;
pub const JS_ATOM_symbol: _bindgen_ty_2 = 75;
pub const JS_ATOM_integer: _bindgen_ty_2 = 76;
pub const JS_ATOM_unknown: _bindgen_ty_2 = 77;
pub const JS_ATOM_arguments: _bindgen_ty_2 = 78;
pub const JS_ATOM_callee: _bindgen_ty_2 = 79;
pub const JS_ATOM_caller: _bindgen_ty_2 = 80;
pub const JS_ATOM__eval_: _bindgen_ty_2 = 81;
pub const JS_ATOM__ret_: _bindgen_ty_2 = 82;
pub const JS_ATOM__var_: _bindgen_ty_2 = 83;
pub const JS_ATOM__arg_var_: _bindgen_ty_2 = 84;
pub const JS_ATOM__with_: _bindgen_ty_2 = 85;
pub const JS_ATOM_lastIndex: _bindgen_ty_2 = 86;
pub const JS_ATOM_target: _bindgen_ty_2 = 87;
pub const JS_ATOM_index: _bindgen_ty_2 = 88;
pub const JS_ATOM_input: _bindgen_ty_2 = 89;
pub const JS_ATOM_defineProperties: _bindgen_ty_2 = 90;
pub const JS_ATOM_apply: _bindgen_ty_2 = 91;
pub const JS_ATOM_join: _bindgen_ty_2 = 92;
pub const JS_ATOM_concat: _bindgen_ty_2 = 93;
pub const JS_ATOM_split: _bindgen_ty_2 = 94;
pub const JS_ATOM_construct: _bindgen_ty_2 = 95;
pub const JS_ATOM_getPrototypeOf: _bindgen_ty_2 = 96;
pub const JS_ATOM_setPrototypeOf: _bindgen_ty_2 = 97;
pub const JS_ATOM_isExtensible: _bindgen_ty_2 = 98;
pub const JS_ATOM_preventExtensions: _bindgen_ty_2 = 99;
pub const JS_ATOM_has: _bindgen_ty_2 = 100;
pub const JS_ATOM_deleteProperty: _bindgen_ty_2 = 101;
pub const JS_ATOM_defineProperty: _bindgen_ty_2 = 102;
pub const JS_ATOM_getOwnPropertyDescriptor: _bindgen_ty_2 = 103;
pub const JS_ATOM_ownKeys: _bindgen_ty_2 = 104;
pub const JS_ATOM_add: _bindgen_ty_2 = 105;
pub const JS_ATOM_done: _bindgen_ty_2 = 106;
pub const JS_ATOM_next: _bindgen_ty_2 = 107;
pub const JS_ATOM_values: _bindgen_ty_2 = 108;
pub const JS_ATOM_source: _bindgen_ty_2 = 109;
pub const JS_ATOM_flags: _bindgen_ty_2 = 110;
pub const JS_ATOM_global: _bindgen_ty_2 = 111;
pub const JS_ATOM_unicode: _bindgen_ty_2 = 112;
pub const JS_ATOM_raw: _bindgen_ty_2 = 113;
pub const JS_ATOM_new_target: _bindgen_ty_2 = 114;
pub const JS_ATOM_this_active_func: _bindgen_ty_2 = 115;
pub const JS_ATOM_home_object: _bindgen_ty_2 = 116;
pub const JS_ATOM_computed_field: _bindgen_ty_2 = 117;
pub const JS_ATOM_static_computed_field: _bindgen_ty_2 = 118;
pub const JS_ATOM_class_fields_init: _bindgen_ty_2 = 119;
pub const JS_ATOM_brand: _bindgen_ty_2 = 120;
pub const JS_ATOM_hash_constructor: _bindgen_ty_2 = 121;
pub const JS_ATOM_as: _bindgen_ty_2 = 122;
pub const JS_ATOM_from: _bindgen_ty_2 = 123;
pub const JS_ATOM_fromAsync: _bindgen_ty_2 = 124;
pub const JS_ATOM_meta: _bindgen_ty_2 = 125;
pub const JS_ATOM__default_: _bindgen_ty_2 = 126;
pub const JS_ATOM__star_: _bindgen_ty_2 = 127;
pub const JS_ATOM_Module: _bindgen_ty_2 = 128;
pub const JS_ATOM_then: _bindgen_ty_2 = 129;
pub const JS_ATOM_resolve: _bindgen_ty_2 = 130;
pub const JS_ATOM_reject: _bindgen_ty_2 = 131;
pub const JS_ATOM_promise: _bindgen_ty_2 = 132;
pub const JS_ATOM_proxy: _bindgen_ty_2 = 133;
pub const JS_ATOM_revoke: _bindgen_ty_2 = 134;
pub const JS_ATOM_async: _bindgen_ty_2 = 135;
pub const JS_ATOM_exec: _bindgen_ty_2 = 136;
pub const JS_ATOM_groups: _bindgen_ty_2 = 137;
pub const JS_ATOM_indices: _bindgen_ty_2 = 138;
pub const JS_ATOM_status: _bindgen_ty_2 = 139;
pub const JS_ATOM_reason: _bindgen_ty_2 = 140;
pub const JS_ATOM_globalThis: _bindgen_ty_2 = 141;
pub const JS_ATOM_bigint: _bindgen_ty_2 = 142;
pub const JS_ATOM_not_equal: _bindgen_ty_2 = 143;
pub const JS_ATOM_timed_out: _bindgen_ty_2 = 144;
pub const JS_ATOM_ok: _bindgen_ty_2 = 145;
pub const JS_ATOM_toJSON: _bindgen_ty_2 = 146;
pub const JS_ATOM_maxByteLength: _bindgen_ty_2 = 147;
pub const JS_ATOM_Object: _bindgen_ty_2 = 148;
pub const JS_ATOM_Array: _bindgen_ty_2 = 149;
pub const JS_ATOM_Error: _bindgen_ty_2 = 150;
pub const JS_ATOM_Number: _bindgen_ty_2 = 151;
pub const JS_ATOM_String: _bindgen_ty_2 = 152;
pub const JS_ATOM_Boolean: _bindgen_ty_2 = 153;
pub const JS_ATOM_Symbol: _bindgen_ty_2 = 154;
pub const JS_ATOM_Arguments: _bindgen_ty_2 = 155;
pub const JS_ATOM_Math: _bindgen_ty_2 = 156;
pub const JS_ATOM_JSON: _bindgen_ty_2 = 157;
pub const JS_ATOM_Date: _bindgen_ty_2 = 158;
pub const JS_ATOM_Function: _bindgen_ty_2 = 159;
pub const JS_ATOM_GeneratorFunction: _bindgen_ty_2 = 160;
pub const JS_ATOM_ForInIterator: _bindgen_ty_2 = 161;
pub const JS_ATOM_RegExp: _bindgen_ty_2 = 162;
pub const JS_ATOM_ArrayBuffer: _bindgen_ty_2 = 163;
pub const JS_ATOM_SharedArrayBuffer: _bindgen_ty_2 = 164;
pub const JS_ATOM_Uint8ClampedArray: _bindgen_ty_2 = 165;
pub const JS_ATOM_Int8Array: _bindgen_ty_2 = 166;
pub const JS_ATOM_Uint8Array: _bindgen_ty_2 = 167;
pub const JS_ATOM_Int16Array: _bindgen_ty_2 = 168;
pub const JS_ATOM_Uint16Array: _bindgen_ty_2 = 169;
pub const JS_ATOM_Int32Array: _bindgen_ty_2 = 170;
pub const JS_ATOM_Uint32Array: _bindgen_ty_2 = 171;
pub const JS_ATOM_BigInt64Array: _bindgen_ty_2 = 172;
pub const JS_ATOM_BigUint64Array: _bindgen_ty_2 = 173;
pub const JS_ATOM_Float16Array: _bindgen_ty_2 = 174;
pub const JS_ATOM_Float32Array: _bindgen_ty_2 = 175;
pub const JS_ATOM_Float64Array: _bindgen_ty_2 = 176;
pub const JS_ATOM_DataView: _bindgen_ty_2 = 177;
pub const JS_ATOM_BigInt: _bindgen_ty_2 = 178;
pub const JS_ATOM_WeakRef: _bindgen_ty_2 = 179;
pub const JS_ATOM_FinalizationRegistry: _bindgen_ty_2 = 180;
pub const JS_ATOM_Map: _bindgen_ty_2 = 181;
pub const JS_ATOM_Set: _bindgen_ty_2 = 182;
pub const JS_ATOM_WeakMap: _bindgen_ty_2 = 183;
pub const JS_ATOM_WeakSet: _bindgen_ty_2 = 184;
pub const JS_ATOM_Iterator: _bindgen_ty_2 = 185;
pub const JS_ATOM_IteratorConcat: _bindgen_ty_2 = 186;
pub const JS_ATOM_IteratorHelper: _bindgen_ty_2 = 187;
pub const JS_ATOM_IteratorWrap: _bindgen_ty_2 = 188;
pub const JS_ATOM_Map_Iterator: _bindgen_ty_2 = 189;
pub const JS_ATOM_Set_Iterator: _bindgen_ty_2 = 190;
pub const JS_ATOM_Array_Iterator: _bindgen_ty_2 = 191;
pub const JS_ATOM_String_Iterator: _bindgen_ty_2 = 192;
pub const JS_ATOM_RegExp_String_Iterator: _bindgen_ty_2 = 193;
pub const JS_ATOM_Generator: _bindgen_ty_2 = 194;
pub const JS_ATOM_Proxy: _bindgen_ty_2 = 195;
pub const JS_ATOM_Promise: _bindgen_ty_2 = 196;
pub const JS_ATOM_PromiseResolveFunction: _bindgen_ty_2 = 197;
pub const JS_ATOM_PromiseRejectFunction: _bindgen_ty_2 = 198;
pub const JS_ATOM_AsyncFunction: _bindgen_ty_2 = 199;
pub const JS_ATOM_AsyncFunctionResolve: _bindgen_ty_2 = 200;
pub const JS_ATOM_AsyncFunctionReject: _bindgen_ty_2 = 201;
pub const JS_ATOM_AsyncGeneratorFunction: _bindgen_ty_2 = 202;
pub const JS_ATOM_AsyncGenerator: _bindgen_ty_2 = 203;
pub const JS_ATOM_EvalError: _bindgen_ty_2 = 204;
pub const JS_ATOM_RangeError: _bindgen_ty_2 = 205;
pub const JS_ATOM_ReferenceError: _bindgen_ty_2 = 206;
pub const JS_ATOM_SyntaxError: _bindgen_ty_2 = 207;
pub const JS_ATOM_TypeError: _bindgen_ty_2 = 208;
pub const JS_ATOM_URIError: _bindgen_ty_2 = 209;
pub const JS_ATOM_InternalError: _bindgen_ty_2 = 210;
pub const JS_ATOM_DOMException: _bindgen_ty_2 = 211;
pub const JS_ATOM_CallSite: _bindgen_ty_2 = 212;
pub const JS_ATOM_Private_brand: _bindgen_ty_2 = 213;
pub const JS_ATOM_Symbol_toPrimitive: _bindgen_ty_2 = 214;
pub const JS_ATOM_Symbol_iterator: _bindgen_ty_2 = 215;
pub const JS_ATOM_Symbol_match: _bindgen_ty_2 = 216;
pub const JS_ATOM_Symbol_matchAll: _bindgen_ty_2 = 217;
pub const JS_ATOM_Symbol_replace: _bindgen_ty_2 = 218;
pub const JS_ATOM_Symbol_search: _bindgen_ty_2 = 219;
pub const JS_ATOM_Symbol_split: _bindgen_ty_2 = 220;
pub const JS_ATOM_Symbol_toStringTag: _bindgen_ty_2 = 221;
pub const JS_ATOM_Symbol_isConcatSpreadable: _bindgen_ty_2 = 222;
pub const JS_ATOM_Symbol_hasInstance: _bindgen_ty_2 = 223;
pub const JS_ATOM_Symbol_species: _bindgen_ty_2 = 224;
pub const JS_ATOM_Symbol_unscopables: _bindgen_ty_2 = 225;
pub const JS_ATOM_Symbol_asyncIterator: _bindgen_ty_2 = 226;
pub const JS_ATOM_END: _bindgen_ty_2 = 227;
pub type _bindgen_ty_2 = ::core::ffi::c_uint;
